import numpy as np
import matplotlib.pyplot as plt
import xarray as xr
import argparse
import atexit
import os
import skamf
import textwrap
from itertools import product


def save():
    if isinstance(amf, dict):
        for a in amf:
            name = amf[a].attrs.pop('filename')
            if name != 'VECTOR_AMF_NO2.nc':
                amf[a].to_netcdf(name)
    else:
        name = amf.attrs.pop('filename')
        if name != 'VECTOR_AMF_NO2.nc':
            amf.to_netcdf(name)
    atexit.unregister(close)
    exit()


def discard():
    atexit.unregister(close)
    exit()


def close():
    inp = ''
    while inp.lower() != 'y' and inp.lower() != 'n':
        inp = input('Save changes [y/n]? ')
        if inp.lower() == 'y':
            save()
        elif inp.lower() == 'n':
            discard()


def _mean(amf: xr.Dataset):
    ret = amf.copy()
    for name in ['clear_box_amf', 'cloudy_box_amf']:
        if name in amf.variables:
            dims = list(amf[name].dims)
            if 'trials' in dims:
                dims.remove('trials')
                ret[name] = (dims, amf[name].mean(dim='trials'))
            std = f'{name}_std'
            if std in amf.variables:
                if amf[name].dims != amf[std].dims:
                    print('standard deviation dataarray did not share dimensions with the amf dataarray')
                    continue
                if 'trials' in amf[std].dims:
                    n = None
                    for var in amf.variables:
                        if 'sasktran_option' in amf[var].attrs:
                            if amf[var].attrs['sasktran_option'] == 'mc: setnumphotonsperlos':
                                n = amf[var].item()  # num photons per trials
                    if n is None:
                        print('unable to find number of photons per trial')
                        continue
                    N = amf.dims['trials']  # num trials
                    ret[std] = (dims, np.sqrt((n * amf[std]**2 + (amf[name] - ret[name])**2).mean(dim='trials') / (n * N)))
    return ret


def _match(*amfs, dataname=None, std=False, **kwargs):

    dnames = ['clear_box_amf', 'cloudy_box_amf', 'clear_box_amf_std', 'cloudy_box_amf_std']
    if dataname is None:
        for name in dnames:
            if all([name in amf.variables for amf in amfs]):
                dataname = name
                break
        if dataname is None:
            print('no valid datanames were found')
            return [None] * len(amfs)
    else:
        if dataname not in dnames:
            print('invalid dataname {dataname}')
            return [None] * len(amfs)

    amfs = list(amfs)
    for i in range(len(amfs)):
        if 'trials' in amfs[i].dims:
            amfs[i] = _mean(amfs[i])

    cs = [set() for a in amfs]
    for i in range(len(amfs)):
        for coord in amfs[i][dataname].coords:
            if not any([n in amfs[i][coord].attrs for n in ['sasktran_option', 'HSP_atmosphere', 'static_atmosphere']]):
                cs[i].add(coord)

    total = set()
    for c in cs:
        total = total | c
    shared = total.copy()
    for c in cs:
        shared = shared & c

    if total != shared:
        print(f'unmatched coordinates {list(total - shared)}')
        return

    for coord in total:
        common = set(np.atleast_1d(amfs[0][coord].values))
        for amf in amfs[1:]:
            common = common & set(np.atleast_1d(amf[coord].values))
        if not common:
            print(f'no common values in coordinate {coord}')
            return
        common = common.pop() if len(common) == 1 else sorted(list(common))
        for i in range(len(amfs)):
            if amfs[i][coord].dims:
                amfs[i] = amfs[i].sel(**{coord: np.atleast_1d(common)})

    for key, value in kwargs.items():
        for i in range(len(amfs)):
            if key in amfs[i].coords:
                amfs[i] = amfs[i].sel(**{key: np.atleast_1d(value)})

    if std:
        dataname = f'{dataname}_std'
        return [amf[dataname] if dataname in amf else None for amf in amfs]
    else:
        return [amf[dataname] for amf in amfs]


def _diff(amf1: xr.Dataset, amf2: xr.Dataset, dataname=None, order=None, percent=True, **kwargs):

    amf1, amf2 = _match(amf1, amf2, dataname=dataname, **kwargs)
    if amf1 is None or amf2 is None:
        return [None] * 4

    if percent:
        diff = 1e2 * (amf2 - amf1) / amf1
    else:
        diff = amf2 - amf1

    # create the order the dimensions are to be stacked in (used by compare())
    dims = sorted(list((set(amf1.coords) | set(amf2.coords)) - {'box_altitude'}))  # these should be the same
    if isinstance(order, str):
        if order not in dims:
            print(f'unknown dimension {order}')
            order = dims.copy()
        else:
            temp = dims.copy()
            temp.remove(order)
            order = [order] + temp
    elif isinstance(order, list):
        temp = []
        for dim in order:
            if dim not in dims:
                print(f'unknown dimension {dim}')
            else:
                temp.append(dim)
        order = temp.copy()
        temp = dims.copy()
        for dim in order:
            temp.remove(dim)
        order += temp
    else:
        order = dims.copy()

    # remove any missing dimensions
    missing = []
    for dim in order:
        if dim not in list(diff.dims):
            missing.append(dim)
    for dim in missing:
        order.remove(dim)

    return amf1, amf2, diff, order


def compare(*amfs, dataname=None, order=None, vmax=None, percent=True, **kwargs):
    """
    Function to compare two AMF tables. Box altitudes are displayed on the y-axis, and all other parameters
    are stacked and displayed on the x axis. The percent difference 100 * (amf2 - amf1) / amf1 is shown.

    Parameters
    ----------
    *amfs : List[xr.Dataset]
        List of amf tables. Only the first two entries will be considered and an error is raised if the length is less
        than 2. Differences will be computed with respect to the first table.
    dataname : str
        The name of the data variable in amf1 and amf2 to compare.
    order : str or List[str] or None
        The order to display the stacked parameters in. Missing parameters will be added in the default order:
        azimuth_difference, solar_zenith, surface_pressure, surface_albedo, viewing_zenith.
    vmax : float or None
        The maximum value to display. Defaults to None, which lets the plot choose vmax.
    **kwargs
        Passed to xr.Dataset.sel to reduce the dataset.
    """

    if len(amfs) < 2:
        print('Two amf tables are required to compare')
        return
    if len(amfs) > 2:
        print('Only the first two amf tables will be considered for this comparison')
    if not all([isinstance(a, xr.Dataset) for a in amfs]):
        print('objects in *amf must be xarray Datasets')
        return

    name1 = amfs[0].attrs['filename']
    name2 = amfs[1].attrs['filename']

    amf1, amf2, diff, order = _diff(amfs[0], amfs[1], dataname, order, percent, **kwargs)
    if amf1 is None or amf2 is None:
        return

    # record the dimensions
    ldims = [len(diff[dim]) for dim in order]
    vdims = [diff[dim].values for dim in order]
    ndims = len(vdims)
    n = np.product(ldims)

    size = np.product([diff[name].size for name in order])
    plt.figure()
    if size > 1:
        # stack the dataset
        stacked = diff.stack(x=tuple(order))

        # create the 2D plot
        data = np.ma.masked_invalid(stacked.values)
        x = np.arange(n + 1)
        y = stacked.box_altitude.values / 1e3
        y -= np.min(y)
        dy = y[1] - y[0]
        y = np.append(y, y[-1] + dy)
        if vmax is None:
            vmin, vmax = -np.max(np.abs(data)), np.max(np.abs(data))
        else:
            vmin = -vmax
        cmap = plt.cm.get_cmap('seismic')
        cmap.set_bad(color=[0.8] * 3)
        plt.pcolormesh(x, y, data, vmin=vmin, vmax=vmax, cmap=cmap)
        plt.colorbar(label='Percent Difference')

        ticks, labels = plt.yticks()
        ticks = list(ticks[ticks >= 0])
        labels = [str(tick) for tick in ticks]

        def label(dim, value):
            if dim == 'surface_albedo':
                return [f'$a_s$', f': {value}']
            elif dim == 'surface_pressure':
                return [f'$P_s$', f': {np.array([v/1e3 for v in value])} kPa']
            elif dim == 'solar_zenith':
                return [f'$\\theta_s$', f': {value}$^\\circ$']
            elif dim == 'viewing_zenith':
                return [f'$\\theta_v$', f': {value}$^\\circ$']
            elif dim == 'azimuth_difference':
                return [f'$\\Delta\\phi$', f': {value}$^\\circ$']
            elif dim == 'cloud_top_pressure':
                return [f'$P_c$', f': {np.array([v/1e3 for v in value])} kPa']
            else:
                return ['---', '---']

        # create the dimension bars
        plt.plot([0, n], [-dy / 2, -dy / 2], 'k--')
        xlabel = []
        for i in range(ndims):
            dbar1 = np.remainder(np.floor(np.arange(n).reshape(1, -1) / np.product(ldims[i + 1:])), ldims[i]) \
                / (ldims[i] - 1)
            plt.pcolormesh(x, np.array([-dy * (ndims - i + 1), -dy * (ndims - i)]), dbar1, vmin=0, vmax=1,
                           cmap='binary')
            ticks.append(-dy * (ndims - i + 0.5))
            sym, val = label(order[i], vdims[i])
            labels.append(sym)
            xlabel.append(f'{order[i]} ({sym}): {val}')

        plt.xticks([], [])
        plt.yticks(ticks, labels)
        plt.xlabel('\n'.join(xlabel[::-1]))
        plt.ylabel('Box Altitude (km)')
        plt.suptitle(f'Comparison: {os.path.split(name2)[1]} vs {os.path.split(name1)[1]}')
        plt.show(block=False)
    else:
        plt.plot(amf1[dataname], diff.box_altitude, label=os.path.split(amf1.attrs['filename'])[1])
        plt.plot(amf2[dataname], diff.box_altitude, label=os.path.split(amf2.attrs['filename'])[1])
        plt.ylabel('Box Altitude (km)')
        plt.xlabel('AMFs')
        plt.suptitle(f'Comparison: {os.path.split(name2)[1]} vs {os.path.split(name1)[1]}')
        plt.legend()
        plt.text(0, 10000, '\n'.join(['{:20}: {:7}'.format(c, diff[c].item()) for c in diff.coords if diff[c].size == 1]), horizontalalignment='left', verticalalignment='top', family='monospace')

    plt.show(block=False)

    return diff


def hist(*amfs, dataname=None, order=None, percent=True, bin_width=None, **kwargs):
    """
    Function to compare two AMF tables. Percent (or absolute) differences are binned and displayed in a histogram.

    Parameters
    ----------
    *amfs : List[xr.Dataset]
        List of amf tables. Only the first two entries will be considered and an error is raised if the length is less
        than 2. Differences will be computed with respect to the first table.
    order : str or List[str] or None
        The order to display the stacked parameters in. Missing parameters will be added in the default order:
        azimuth_difference, solar_zenith, surface_pressure, surface_albedo, viewing_zenith.
    vmax : float or None
        The maximum value to display. Defaults to None, which lets the plot choose vmax.
    **kwargs
        Passed to xr.Dataset.sel to reduce the dataset.
    """

    if len(amfs) < 2:
        print('Two amf tables are required to compare')
        return
    if len(amfs) > 2:
        print('Only the first two amf tables will be considered for this comparison')
    if not all([isinstance(a, xr.Dataset) for a in amfs]):
        print('objects in *amf must be xarray Datasets')
        return

    amf1, amf2, diff, order = _diff(amfs[0], amfs[1], dataname, order, percent, **kwargs)
    if amf1 is None or amf2 is None:
        return

    plt.figure()
    bin_width = 2 if percent else 0.2
    bins = None if bin_width is None else np.arange(
        np.floor(diff.min() / bin_width) * bin_width,
        np.ceil(diff.max() / bin_width) * bin_width + bin_width,
        bin_width
    )
    data = diff.values.flatten()
    data = data[np.isfinite(data)]
    # data[~np.isfinite(data)] = np.nan
    plt.hist(data, bins=bins)
    if percent:
        plt.xlabel('Percent Difference')
    else:
        plt.xlabel('Difference')
    plt.text(diff.max().item() / 2, diff.size / 10, 'Average: {:.2f}'.format(np.abs(diff).mean().item()))
    plt.show(block=False)

    return amf1, amf2


def scatter(*amfs, dataname=None, order=None, highlight=None, title=None, xlabel=None, ylabel=None, **kwargs):
    """
    Function to compare two AMF tables. The first table is placed on the x axis and the second on the y.

    Parameters
    ---------
    *amfs : List[xr.Dataset]
        List of amf tables. Only the first two entries will be considered and an error is raised if the length is less
        than 2. Differences will be computed with respect to the first table.
    order : str or List[str] or None
        The order to display the stacked parameters in. Missing parameters will be added in the default order:
        azimuth_difference, solar_zenith, surface_pressure, surface_albedo, viewing_zenith.
    hightlight : dict
        dictionary containing coords and values {coord: values} to be highlighted.
    **kwargs
        Passed to xr.Dataset.sel to reduce the dataset.
    """

    if len(amfs) < 2:
        print('Two amf tables are required to compare')
        return
    if len(amfs) > 2:
        print('Only the first two amf tables will be considered for this comparison')
    if not all([isinstance(a, xr.Dataset) for a in amfs]):
        print('objects in *amf must be xarray Datasets')
        return

    name1 = amfs[0].attrs['filename']
    name2 = amfs[1].attrs['filename']

    amf1, amf2, diff, order = _diff(amfs[0], amfs[1], dataname, order, True, **kwargs)
    if amf1 is None or amf2 is None:
        return
    order += ['box_altitude']

    plt.figure()
    highlight = {} if highlight is None else highlight
    if len(highlight) > 0:
        lowlight = {name: diff[name].values for name in order}
        for key, value in highlight.items():
            new_values = list(set(np.atleast_1d(lowlight[key])) - set(np.atleast_1d(highlight[key])))
            lowlight[key] = np.atleast_1d(new_values)
            highlight[key] = np.atleast_1d(value)

        def array_to_str(arr, fmt):
            arrstr = [fmt.format(a) for a in arr]
            return ','.join(arrstr)

        xhigh = amf1.sel(**highlight).stack(x=tuple(order)).values
        yhigh = amf2.sel(**highlight).stack(x=tuple(order)).values

        xlow = amf1.sel(**lowlight).stack(x=tuple(order)).values
        ylow = amf2.sel(**lowlight).stack(x=tuple(order)).values

        small = min(np.min(xhigh), np.min(yhigh), np.min(xlow), np.min(ylow))
        big = min(np.max(xhigh), np.max(yhigh), np.max(xlow), np.max(ylow))

        plt.plot(xlow, ylow, 'kx', label=f'{key}={array_to_str(lowlight[key], "{:.1f}")}')
        plt.plot(xhigh, yhigh, 'gx', label=f'{key}={array_to_str(highlight[key], "{:.1f}")}')

        plt.plot([small, big], [small, big], 'b')
        plt.legend()
    else:
        x = amf1.stack(x=tuple(order)).values
        y = amf2.stack(x=tuple(order)).values

        small = min(np.min(x), np.min(y))
        big = min(np.max(x), np.max(y))

        plt.plot(x, y, 'kx')
        plt.plot([small, big], [small, big], 'b')

    title = 'AMF Comparison' if title is None else title
    xlabel = f'AMFs: {os.path.split(name1)[1]}' if xlabel is None else xlabel
    ylabel = f'AMFs: {os.path.split(name2)[1]}' if ylabel is None else ylabel
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    plt.show(block=False)

    return diff


def plot(*amfs: xr.Dataset, dataname=None, reference=None, **kwargs):
    if len(amfs) < 1:
        print('At least one amf table is required to plot')
        return
    if reference is None and len(amfs) > 4:
        print('Only the first four amf tables will be considered')
    if reference is not None and len(amfs) < 2:
        print('At least two tables are required if a reference table is specified')
        return
    if reference is not None and len(amfs) > 5:
        print('Only the first five amf tables will be considered')
    if not all([isinstance(a, xr.Dataset) for a in amfs]):
        print('objects in *amf must be xarray Datasets')
        return
    names = [a.attrs['filename'] for a in amfs]
    stds = _match(*amfs, dataname=dataname, std=True, **kwargs)
    amfs = _match(*amfs, dataname=dataname, **kwargs)
    if amfs[0] is None:
        return

    if reference is not None:
        i = list(range(len(amfs)))
        if reference not in i:
            print(f'invalid reference index {reference}')
            return
        else:
            i.remove(reference)
            stds = [stds[j] - stds[reference] if stds[j] is not None and stds[reference] is not None else None for j in i]
            amfs = [amfs[j] - amfs[reference] for j in i]
            names = [names[j] for j in i]

    plt.figure()
    if reference is not None:
        plt.plot([0, 0], [amfs[0].box_altitude.values[0], amfs[0].box_altitude.values[-1]], 'k--')
    coords = sorted(list(set(amfs[0].coords) - {'box_altitude'}))
    values = [amfs[0][c].values for c in coords]
    l = np.array([len(v) for v in values])
    linestyles = ['-', '--', '-.', ':']
    for i, comb in enumerate(product(*values)):
        data = [a.sel(**{c: v for c, v in zip(coords, comb)}).values for a in amfs]
        error = [s.sel(**{c: v for c, v in zip(coords, comb)}).values if s is not None else None for s in stds]
        for d, e, line in zip(data, error, linestyles):
            plt.plot(d, amfs[0].box_altitude.values, color=f'C{i%10}', linestyle=line,
                     label=dict(zip(np.array(coords)[l > 1], np.array(comb)[l > 1])))
            if e is not None:
                plt.fill_betweenx(amfs[0].box_altitude.values, d - e, d + e, color=f'C{i%10}', alpha=0.2)
    handles, labels = plt.gca().get_legend_handles_labels()
    import matplotlib.lines as mlines

    n = len(amfs)
    handles = handles[::n] + [mlines.Line2D([], [], color='k', linestyle=l) for l in linestyles]
    labels = labels[::n] + [os.path.basename(n) for n in names]
    title = f'AMFs: {dict(zip(np.array(coords)[l==1], np.array([v[0] for v in np.array(values)[l==1]])))}'
    plt.title(textwrap.fill(title, 80))
    plt.legend(handles, labels)
    plt.xlabel('AMF')
    plt.ylabel('Altitude (m)')

    plt.show(block=False)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('filename', nargs='+')
    parser.add_argument('--ec', action='store_true')
    args = vars(parser.parse_args())

    amf = {}

    if args['ec']:
        amf['EC'] = xr.open_dataset(os.path.join(skamf.__path__[0], 'VECTOR_AMF\\VECTOR_AMF_NO2.nc'))
        amf['EC'] = amf[0].load()
        amf['EC'].close()
        amf['EC'].attrs['filename'] = os.path.join(skamf.__path__[0], 'VECTOR_AMF\\VECTOR_AMF_NO2.nc')

    for filename in args['filename']:
        if os.path.isfile(filename) and filename[-3:] == '.nc':
            name = os.path.splitext(os.path.basename(filename))[0]
            amf[name] = xr.open_dataset(filename)
            amf[name] = amf[name].load()
            amf[name].close()
            amf[name].attrs['filename'] = filename
        elif os.path.isdir(filename):
            for root, dirs, files in os.walk(filename):
                for filename in files:
                    if filename[-3:] == '.nc':
                        name = os.path.splitext(filename)[0]
                        amf[name] = xr.open_dataset(os.path.join(root, filename))
                        amf[name] = amf[name].load()
                        amf[name].close()
                        amf[name].attrs['filename'] = os.path.join(root, filename)

    if len(amf) == 1:
        amf = amf.popitem()[1]

    atexit.register(close)