import numpy as np
import xarray as xr
import sasktran as sk
import logging
import doas.amf.amf
from skamf.atmosphere import hsp_to_sasktran, is_hsp_atmosphere, is_static_atmosphere, static_to_sasktran
from skamf.utility import parse_option_names

log = logging.getLogger()


"""
Hard-coded settings
"""
wavelength = 440
latitude = 57.1
longitude = 248.4
time = 58197.666

cloudy_case_surface_pressure = 100000

onced_warnings = set({})
onced_debug = set()
onced_info = set()


def warn_once(msg: str):
    h = hash(msg)
    if h not in onced_warnings:
        log.warning(msg)
        onced_warnings.add(h)


def debug_once(msg: str):
    h = hash(msg)
    if h not in onced_debug:
        log.debug(msg)
        onced_debug.add(h)


def info_once(msg: str):
    h = hash(msg)
    if h not in onced_info:
        log.info(msg)
        onced_info.add(h)


class AMFBase:
    def __init__(self, table_subset: xr.Dataset, engine: str):
        self.amf_table = table_subset.copy()  # new view
        self.used = set({})
        self._amf_engine = None
        self._atmosphere = None
        self._geometry = None
        self._geometry_coords = {}
        self._los_coord_names = []
        self._shell_spacing = 0.0
        self._fixed_boxes = 0
        self._surface_altitude = None
        self._engine = engine
        self._layers_above_cloud = True
        self._cloud_below_ground = False
        self._cloud_top_altitude = None
        debug_once(f"Calculating AMF's at {wavelength} [nm]")
        debug_once(f"Calculating AMF's with a reference point of ({latitude} [degrees_north], {longitude} [degrees_east])")
        debug_once(f"Calculating AMF's with a reference date of {time} [days since 1858-11-17 0:0:0]")
        self._make_atmosphere()
        self._make_geometry()
        self._mark_engine_options()
        self._make_amf_engine()
        self._check_used()

    def __getitem__(self, item: str):
        try:
            value = self.amf_table.cf[item]
            self.used.add(value.name)
            return value
        except KeyError:
            return None

    def __setitem__(self, key, value):
        raise NotImplementedError()

    class TemporaryOption:
        def __init__(self, outer_amfbase_instance, **options):
            self._table_options = outer_amfbase_instance.amf_table.options
            self._new_options = options
            self._old_options = {}

        def __enter__(self):
            for key, value in self._new_options.items():
                if key in self._table_options:
                    self._old_options[key] = self._table_options.pop(key)
                self._table_options[key] = value

        def __exit__(self):
            # remove the temporary options
            for key, value in self._new_options.items():
                self._table_options.pop(key)
            # restore the original options
            for key, value in self._old_options.items():
                self._table_options[key] = value

    def assign_temporary_option(self, **options):
        return AMFBase.TemporaryOption(self, **options)

    def get_engine_options(self, engine_name: str):
        options = {}
        for var in self.amf_table.variables:
            option_name = parse_option_names(self.amf_table, var).get(engine_name, None)
            if option_name is not None:
                # get variable as scalar or 1d array
                option = self.amf_table[var]
                if len(np.atleast_1d(option)) == 1:
                    option = np.asscalar(option)
                else:
                    option = option.values.copy()

                # check that this option name isn't already present
                if option_name in options:
                    info_once(f"Multiple instances of engine option '{option_name}'")
                else:
                    options[option_name.lower()] = option

        if 'manualdiffuseheights' in options:
            surface = options.get('surfaceheight', 0.0)
            toa = options.get('toaheight', 1e5)
            heights = options['manualdiffuseheights']
            heights = np.array([surface + 1e-6, *heights[heights > surface + 1]])
            heights = heights[heights < toa]
            if heights != options['manualdiffuseheights']:
                info_once('manualdiffuseheights below the surface height {:.1f} m and above the top-of-atmosphere {:.1f} m were truncated, and a height just above the surface was added'.format(surface, toa))
            options['manualdiffuseheights'] = heights

        for opt in ['manualraytracingshells', 'manualsolarraytracingshells']:
            if opt in options:
                surface = options.get('surfaceheight', 0.0)
                toa = options.get('toaheight', 1e5)
                heights = options[opt]
                heights[heights < surface] = surface
                heights[heights > toa] = toa
                heights = np.unique(heights)
                if heights != options[opt]:
                    info_once('{} below the surface height {:.1f} m and above the top-of-atmosphere {:.1f} m were truncated'.format(opt, surface, toa))
                options[opt] = heights

        # set defaults if they are not set already
        defaults = self._default_engine_options(engine_name)
        for opt, value in defaults.items():
            if opt not in options:
                options[opt] = value

                info_once(f'{engine_name} option {opt} was not found in the user-specified options so the default setting was used')

        return options

    def _default_engine_options(self, engine_name: str):
        # default engine options suitable for AMF calculation
        if engine_name == 'do' or engine_name == 'disco':
            # good improvement was seen from 16 to 20 streams, but limited from 20 to 24
            # good improvement was seen from 50 to 200 to 300 layers, but limited from 300 to 500
            return {'numlayers': 300, 'numstreams': 20, 'averagereferencepoint': True}
        elif engine_name == 'hr':
            # these settings were found to prevent unphysical oscillations from amf layers not aligning with settings
            amf_bot = self.amf_engine.box_boundaries[0]
            amf_top = self.amf_engine.box_boundaries[-1]
            amf_del = self._shell_spacing
            atmo_bot = self['surface_altitude'].item()
            atmo_top = self['altitude_at_top_of_atmosphere_model'].item()

            # one diffuse height at 1e-6m above the surface, two in each amf layer, one every km outside the layers
            amf_diffuse = np.arange(amf_bot + 0.25 * amf_del, amf_top, 0.5 * amf_del)
            atmo_diffuse = np.arange(5e2, atmo_top, 1e3)
            diffuse = np.unique([atmo_bot, *amf_diffuse, *atmo_diffuse[np.logical_or(atmo_diffuse < amf_bot, atmo_diffuse > amf_top)]])

            # two ray tracing / solar ray tracing shells per amf layer, 1 km shells outside the layers
            amf_shells = np.arange(amf_bot, amf_top + 1, 0.5 * amf_del)
            atmo_shells = np.arange(atmo_bot, atmo_top, 1e3)
            shells = np.unique([*amf_shells, *atmo_shells[np.logical_or(atmo_shells < amf_bot, atmo_shells > amf_top)], atmo_top])

            # four optical heights per amf layer, one every 500 m outside the layers
            amf_optical = np.arange(amf_bot, amf_top + 1, 0.25 * amf_del)
            atmo_optical = np.arange(atmo_bot, atmo_top, 5e2)
            optical = np.unique([*amf_optical, *atmo_optical[np.logical_or(atmo_optical < amf_bot, atmo_optical > amf_top)], atmo_top])

            return {'manualdiffuseheights': diffuse, 'manualraytracingshells': shells,
                    'manualsolarraytracingshells': shells, 'manualopticalheights': optical}
        elif engine_name == 'mc':
            # running up to a target accuracy is not supported for MC AMFs, so set an unreachable target and set the number of photons
            # 1e7 photons was found to produce very nice AMFs but it is very costly - 1e6 is a reasonable starting point
            return {'settargetstd': 0, 'setnumphotonsperlos': 1e6}
        else:
            log.error(f'Unknown engine name {engine_name}')

    def mark_as_engine_option(self, var_name, disco: str=None, hr: str=None, mc: str=None):
        option_str = ''
        if disco is not None:
            option_str += f'disco: {disco} '
        if hr is not None:
            option_str += f'hr: {hr} '
        if mc is not None:
            option_str += f'mc: {mc} '
        self.amf_table[var_name].attrs['sasktran_option'] = option_str

    @property
    def surface_pressure(self) -> float:
        p_surface = self['surface_air_pressure']
        if p_surface is None:
            warn_once('No surface pressure was given (surface_air_pressure). Using a default value of 100 kPa')
            p_surface = 100000
        else:
            p_surface = p_surface.item()
        return p_surface

    @property
    def surface_albedo(self) -> float:
        raise NotImplementedError()

    @property
    def atmosphere(self) -> sk.Atmosphere:
        return self._atmosphere

    @property
    def geometry(self) -> sk.Geometry:
        return self._geometry

    @property
    def amf_engine(self) -> doas.amf.amf.AirMassFactorBase:
        return self._amf_engine

    def pressure_to_altitude(self, pressure_Pa):
        return np.interp(np.log(pressure_Pa), np.log(self._pressure_interp_pressure[::-1]), self._pressure_interp_altitude[::-1])

    def _make_atmosphere(self):
        if is_hsp_atmosphere(self.amf_table):
            atmo, info = hsp_to_sasktran(self.amf_table, self.surface_pressure, 0.0)
        elif is_static_atmosphere(self.amf_table):
            atmo, info = static_to_sasktran(self.amf_table, self.surface_pressure, 0.0)
        else:
            raise ValueError('No valid atmosphere found in AMF table')
        self._pressure_interp_pressure = info['pressure_edges']
        self._pressure_interp_altitude = info['altitude_edges']
        atmo.brdf = self.surface_albedo

        # set the top of atmosphere
        self.amf_table['top_of_atmosphere'] = info['top_of_atmosphere_altitude']
        self.amf_table['top_of_atmosphere'].attrs['units'] = 'm'
        self.amf_table['top_of_atmosphere'].attrs['standard_name'] = 'altitude_at_top_of_atmosphere_model'

        # set the surface altitude
        self.amf_table['surface_altitude'] = info['surface_altitude']
        self.amf_table['surface_altitude'].attrs['units'] = 'm'
        self.amf_table['surface_altitude'].attrs['standard_name'] = 'surface_altitude'

        # set the altitude grid
        # grid = np.append(info['altitude'], np.linspace(0, 1e5, 401))
        # grid[grid > info['top_of_atmosphere_altitude']] = info['top_of_atmosphere_altitude']
        # grid[grid < info['surface_altitude']] = info['surface_altitude']
        # grid = np.unique(grid)
        # grid = np.unique(10*np.round(grid/10))  # round to nearest 10 to avoid tiny layers
        grid = np.linspace(0, 1e5, 401)
        self.amf_table['altitude_grid'] = xr.DataArray(grid, dims=['z'])

        # set the atmosphere property
        self._atmosphere = atmo

    def _make_geometry(self):
        geometry = sk.NadirGeometry()

        # get the observation zenith angles
        observer_zeniths_name = None
        observer_zeniths = self['platform_zenith_angle']
        if observer_zeniths is None:
            warn_once('No observer zenith angle was given (platform_zenith_angle). '
                      'Using a default value of 60 degrees.')
            observer_zeniths = np.array([60.0])
        else:
            observer_zeniths_name = observer_zeniths.name
            observer_zeniths = np.atleast_1d(observer_zeniths)

        # get the solar zenith angle
        solar_zenith = self['solar_zenith_angle']
        if solar_zenith is None:
            warn_once('No solar zenith angle was given (solar_zenith_angle). '
                      'Using a default value of 60 degrees.')
            solar_zenith = 60.0
        else:
            solar_zenith = solar_zenith.item()

        # get the azimuthal differences
        azimuth_differences_name = None
        azimuth_differences = self['angle_of_rotation_from_solar_azimuth_to_platform_azimuth']
        if azimuth_differences is None:
            warn_once('Differences in azimuth angles were not specified '
                      '(angle_of_rotation_from_solar_azimuth_to_platform_azimuth). '
                      'Using a default value of 60 degrees.')
            azimuth_differences = np.array([60.0])
        else:
            azimuth_differences_name = azimuth_differences.name
            azimuth_differences = np.atleast_1d(azimuth_differences)

        # generate a meshgrid of angles
        observer_zeniths, azimuth_differences = np.meshgrid(observer_zeniths, azimuth_differences)

        # save observer_zeniths and azimuth_differences coordinates
        if observer_zeniths_name is not None:
            self._geometry_coords[observer_zeniths_name] = {'data': observer_zeniths.flatten(), 'dims': ['los']}
            self._los_coord_names.append(observer_zeniths_name)
        if azimuth_differences_name is not None:
            self._geometry_coords[azimuth_differences_name] = {'data': azimuth_differences.flatten(), 'dims': ['los']}
            self._los_coord_names.append(azimuth_differences_name)
        geometry.from_zeniths_and_azimuth_difference(
            solar_zenith, observer_zeniths, azimuth_differences, mjd=time,
            reference_point=(latitude, longitude, self['surface_altitude'].item(), time)
        )

        if self._engine == 'do':
            geometry.reference_point = None

        self._geometry = geometry

    def _mark_engine_options(self):
        raise NotImplementedError()

    def _fix_box_altitudes(self, boundaries, midpoints):
        raise NotImplementedError()

    def _select_amf_engine(self) -> doas.amf.amf.AirMassFactorBase:
        if self._engine == 'hr':
            return doas.AirMassFactorFiniteDifferenceHR(self.geometry, self.atmosphere, 'no2')
        elif self._engine == 'do':
            return doas.AirMassFactorDiscO(self.geometry, self.atmosphere, 'no2')
        elif self._engine == 'mc':
            return doas.AirMassFactorMC(self.geometry, self.atmosphere, 'no2')
        else:
            raise ValueError(f'AMFWrapper: invalid engine {self._engine}')

    def _make_amf_engine(self):
        self._amf_engine = self._select_amf_engine()

        # setup engine options
        options = {}

        # set altitudes
        box_altitudes = np.atleast_1d(self['box_altitude'])
        center_distances = np.diff(box_altitudes)
        mean_distance = np.mean(center_distances)
        std_distance = np.std(center_distances)
        if std_distance > 1e-2:
            log.error('box altitudes are not linearly spaced')
            exit(1)
        box_boundaries = np.array([box_altitudes[0] - mean_distance / 2, *(mean_distance / 2 + box_altitudes)])
        self._shell_spacing = mean_distance

        box_boundaries, box_midpoints = self._fix_box_altitudes(
            self['surface_altitude'].item() + box_boundaries,
            self['surface_altitude'].item() + box_altitudes
        )
        if len(box_midpoints) == 0:
            self._layers_above_cloud = False

        # set AMF engine box boundaries
        self._amf_engine.box_boundaries = box_boundaries
        self._amf_engine.box_midpoints = box_midpoints
        self._amf_engine.wf_widths = 0.5 * (box_boundaries[1:] - box_boundaries[:-1])

        # set options
        options.update(self.get_engine_options(self._engine))
        self._amf_engine.options.update(options)
        debug_once(f'Using SASKTRAN-{self._engine.upper()} with options {set(self._amf_engine.options.keys())}')

    def _check_used(self):
        """
        This function checks that all of the AMF tables coordinates have been referenced (via self[<coord_name>]).
        It provides a safeguard to make sure they are all being used.

        A warning will be logged ONCE if a coordinate is not being used.
        """
        self.ignore_coords()
        unused = (self.amf_table.coords.keys()) - self.used
        unused = [u for u in unused if 'sasktran_option' not in self[u].attrs]
        if len(unused) > 0:
            warn_once(f"The AMF table coordinates {unused} are not being used.")

    def ignore_coords(self):
        raise NotImplementedError()

    def calculate(self):
        """
        Compute the box AMF's for the subset of the AMF table.

        Notes
        -----
        This function calls the AMF engine (i.e. self.amf_engine.__call__) and reformats the coordinates of the
        returned xr.DataArray such that it can be returned to the AMF table (and fluidly merged in).

        Returns
        -------
        xr.DataArray
            The computed box AMF's.
        """
        if not self._layers_above_cloud:
            warn_once('No full AMF layers were found above the cloud top pressure, so no AMFs were calculated.')
            return False
        if self._cloud_below_ground:
            warn_once('Cloud altitude lower than surface altitude, so no AMFs were calculated.')
            return False

        amfs = self.amf_engine([wavelength])
        for name, coord in self._geometry_coords.items():
            amfs.coords[name] = xr.DataArray(**coord)
        amfs = amfs.set_index(los=self._los_coord_names).unstack('los').unstack('perturbation')

        coords = {c for c in amfs.coords if amfs.coords[c].size == 1} - {'altitude'}
        amfs = amfs.squeeze(dim=coords, drop=True)

        std = 'amfstd' in amfs

        altitude = amfs.coords['altitude']
        altitude_above_surface = self['box_altitude'].values[self._fixed_boxes:]
        surface_altitude = self['surface_altitude'].item()
        box_boundaries = self.amf_engine.box_boundaries

        if np.any(altitude > box_boundaries[1:]) or np.any(altitude < box_boundaries[:-1]):  # make sure each AMF calculation is within the correct box
            logging.error('The altitudes where AMFs were calculated do not match the expected altitudes.')

        if self._fixed_boxes > 0:
            # exclude the first point as this will be different if the cloud is in the middle of a layer
            if len(altitude) > 1 and np.max(np.abs(altitude[1:] - altitude_above_surface[1:] - surface_altitude)) > 1.0:  # make sure each AMF calculation altitude matches what is expected
                logging.error('The altitudes where AMFs were calculated do not match the expected altitudes.')

            if np.abs(altitude[0] - altitude_above_surface[0] - surface_altitude) > 1.0:  # if the lowest point is not on the regular spacing (because of a cloud)
                lowest_box = box_boundaries[1] - box_boundaries[0]

                # scale the lowest AMF down according to how much smaller the layer is
                if len(altitude) == 1:
                    pass
                if lowest_box < 0.1 * self._shell_spacing:
                    # if the box is too small, then the wf_width will be small and the result might be invalid
                    # use the result from the layer above
                    amfs['amf'].loc[dict(altitude=altitude[0])] = (lowest_box / self._shell_spacing) * amfs['amf'].loc[dict(altitude=altitude[1])]
                    if std:
                        amfs['amfstd'].loc[dict(altitude=altitude[0])] = (lowest_box / self._shell_spacing) * amfs['amfstd'].loc[dict(altitude=altitude[1])]
                elif lowest_box < self._shell_spacing or len(altitude):
                    amfs['amf'].loc[dict(altitude=altitude[0])] *= lowest_box / self._shell_spacing
                    if std:
                        amfs['amfstd'].loc[dict(altitude=altitude[0])] *= lowest_box / self._shell_spacing
                else:  # should not happen (if the AMF altitude isn't on the regular spacing then the bottom box should be smaller)
                    logging.error('The altitudes where AMFs were calculated do not match the expected altitudes.')

            amfs.coords['altitude'] = self['box_altitude'].values[self._fixed_boxes:]
        else:
            if np.max(np.abs(altitude - altitude_above_surface - surface_altitude)) > 1.0:  # make sure each AMF calculation altitude matches what is expected
                logging.error('The altitudes where AMFs were calculated do not match the expected altitudes.')
            amfs.coords['altitude'] = self['box_altitude'].values
        amfs = amfs.rename({'altitude': 'box_altitude'})

        # drop extra data variables produced by the doas module
        drop_labels = set(amfs.data_vars) - {'amf', 'amfstd', 'total_amf_sum', 'total_amf_calc'}
        amfs = amfs.drop(drop_labels)

        if self._fixed_boxes > 0:
            box_amfs = xr.Dataset()
            for c in amfs['amf'].coords:
                if c != 'box_altitude':
                    box_amfs.coords[c] = amfs[c]
                else:
                    box_amfs[c] = self['box_altitude'][:self._fixed_boxes]

            dims = amfs['amf'].dims
            shape = []
            for d in dims:
                if d != 'box_altitude':
                    shape.append(amfs.dims[d])
                else:
                    shape.append(self._fixed_boxes)
            box_amfs['amf'] = xr.DataArray(np.zeros(shape), dims=dims)
            if std:
                box_amfs['amfstd'] = xr.DataArray(np.zeros(shape), dims=dims)
            con = xr.Dataset(data_vars={'amf': amfs['amf']})
            if std:
                con['amfstd'] = amfs['amfstd']
            box_amfs = xr.concat([box_amfs, con], 'box_altitude', data_vars='minimal', coords='minimal')
        else:
            box_amfs = amfs

        self._box_amf = box_amfs['amf']
        if 'amfstd' in box_amfs:
            self._box_amf_std = box_amfs['amfstd']
        if 'total_amf_sum' in amfs:
            self._total_amf_boxsum = amfs['total_amf_sum']
        if 'total_amf_calc' in amfs:
            self._total_amf_noNO2 = amfs['total_amf_calc']

        return True

    @property
    def box_amf(self):
        return self._box_amf

    @property
    def box_amf_std(self):
        if hasattr(self, '_box_amf_std'):
            return self._box_amf_std
        else:
            return None

    @property
    def amf_box_sum(self):
        if hasattr(self, '_total_amf_boxsum'):
            return self._total_amf_boxsum
        else:
            return None

    @property
    def amf_no_absorber(self):
        if hasattr(self, '_total_amf_noNO2'):
            return self._total_amf_noNO2
        else:
            return None


class ClearSkyAMF(AMFBase):
    def ignore_coords(self):
        pass

    def _mark_engine_options(self):
        # set HR options
        self.mark_as_engine_option(self['surface_altitude'].name, hr='surfaceheight', mc='surfaceheight')
        self.mark_as_engine_option(self['altitude_at_top_of_atmosphere_model'].name, hr='toaheight', mc='toaheight')
        # set Disco options
        self.mark_as_engine_option('altitude_grid', disco='altitudegrid')

    def _fix_box_altitudes(self, boundaries, midpoints):
        return boundaries, midpoints

    @property
    def surface_albedo(self) -> float:
        albedo = self['surface_albedo']
        if albedo is None:
            warn_once('No surface albedo was given (surface_albedo). Using a default value of 0.3.')
            albedo = 0.3
        else:
            albedo = albedo.item()
        return albedo


class CloudySkyAMF(AMFBase):
    def ignore_coords(self):
        pass

    def _mark_engine_options(self):
        cloud_top = self.pressure_to_altitude(self['air_pressure_at_cloud_top'].item())
        self.amf_table['cloud_top_altitude'] = cloud_top

        # set HR options
        self.mark_as_engine_option('cloud_top_altitude', hr='surfaceheight', mc='surfaceheight')
        self.mark_as_engine_option(self['altitude_at_top_of_atmosphere_model'].name, hr='toaheight', mc='toaheight')

        # set Disco options
        altitudegrid = self.amf_table['altitude_grid']
        altitudegrid = np.array([cloud_top, *altitudegrid[np.where(altitudegrid > cloud_top)]])

        self.amf_table = self.amf_table.drop('altitude_grid').squeeze()

        self.amf_table['altitude_grid'] = xr.DataArray(altitudegrid, dims=['z'])
        self.mark_as_engine_option('altitude_grid', disco='altitudegrid')

    def _fix_box_altitudes(self, boundaries, midpoints):
        # get indexes to keep
        self._cloud_top_altitude = self.pressure_to_altitude(self['air_pressure_at_cloud_top'].item())
        if self._cloud_top_altitude >= np.max(boundaries):
            return np.array([]), np.array([])

        mask = np.where(boundaries > self._cloud_top_altitude)

        # get new altitudes
        new_boundaries = np.unique(np.append(boundaries[mask], self._cloud_top_altitude))
        new_midpoints = 0.5 * (new_boundaries[:-1] + new_boundaries[1:])

        # note how many we are chopping off
        self._fixed_boxes = len(midpoints) - len(new_midpoints)

        return new_boundaries, new_midpoints

    @property
    def surface_albedo(self) -> float:
        albedo = self['cloud_albedo']
        if albedo is None:
            warn_once('No cloud albedo was given (surface_albedo). Using a default value of 0.8.')
            albedo = 0.8
        else:
            albedo = albedo.item()
        return albedo

    def _make_amf_engine(self):
        super()._make_amf_engine()
        if self['air_pressure_at_cloud_top'].item() > self.surface_pressure:
            self._cloud_below_ground = True
