"""
This file generates a new AMF table.
"""

import numpy as np
import xarray as xr

from skamf.utility import get_response, parse_response
from skamf.atmosphere import is_hsp_atmosphere, default_hsp_atmosphere
from skamf import _coord_info


def new_table(clear_sky: bool, cloudy_sky: bool, total_sum: bool, total_remove: bool, engine: str,
              **kwargs) -> xr.Dataset:
    print('Welcome to the AMF table wizard.')
    print('You will be prompted for the new AMF tables coordinates.')
    print('Coordinate values should be delimited by spaces or commas.')
    print("Respond with 'omit' to not include a coordinate in the table.")
    print('An empty response will select the default values.')
    print("Enter 'exit' to quit without making the table.")
    print('\n' + '-' * 80)

    table = xr.Dataset()

    # add all relevant coords defined in __init__.py to the new table, request user input where necessary
    for coord, info in _coord_info.items():
        if clear_sky and info['clear_sky'] or cloudy_sky and info['cloudy_sky']:
            if coord in kwargs and kwargs[coord] is not None:
                values = parse_response(
                    response=kwargs[coord],
                    varname=coord,
                    default=info['default'],
                    mandatory=info['mandatory']
                )
            else:
                values = get_response(
                    varname=coord,
                    prompt=info['prompt'],
                    default=info['default'],
                    mandatory=info['mandatory']
                )
            if values is not None:
                table.coords[coord] = values
                table.coords[coord].attrs['units'] = info['units']
                table.coords[coord].attrs['long_name'] = info['long_name']
                table.coords[coord].attrs['standard_name'] = info['standard_name']
                print(table.coords[coord])
            print('\n' + '-' * 80)

    if 'box_altitude' in table.coords:
        table.coords['box_altitude'].attrs['positive'] = 'up'

    # construct the clear AMF table if requested
    if clear_sky:
        clear_coords = [c for c in _coord_info if _coord_info[c]['clear_sky'] and len(table.coords[c].shape) > 0]
        table['clear_box_amf'] = xr.DataArray(
            data=np.zeros([table[c].shape[0] for c in clear_coords]),
            coords=[table[c] for c in clear_coords]
        )
        table['clear_box_amf'].attrs['long_name'] = 'NO2 box air mass factors'
        table['clear_box_amf'].attrs['units'] = '1'

        # add a data variable for uncertainty if this table will be processed with the MC engine
        if engine == 'mc':
            table['clear_box_amf_std'] = xr.DataArray(
                data=np.zeros([table[c].shape[0] for c in clear_coords]),
                coords=[table[c] for c in clear_coords]
            )
            table['clear_box_amf_std'].attrs['long_name'] = 'NO2 box air mass factor standard deviations'
            table['clear_box_amf_std'].attrs['units'] = '1'

        # add data variable for total AMFs if requested
        clear_coords.remove('box_altitude')
        if total_sum:
            table['clear_amf_box_sum'] = xr.DataArray(
                data=np.zeros([table[c].shape[0] for c in clear_coords]),
                coords=[table[c] for c in clear_coords]
            )
        if total_remove:
            table['clear_amf_no_NO2'] = xr.DataArray(
                data=np.zeros([table[c].shape[0] for c in clear_coords]),
                coords=[table[c] for c in clear_coords]
            )
        del clear_coords

    # construct the cloudy AMF table if requested
    if cloudy_sky:
        cloudy_coords = [c for c in _coord_info if _coord_info[c]['cloudy_sky'] and len(table.coords[c].shape) > 0]
        table['cloudy_box_amf'] = xr.DataArray(
            data=np.zeros([table[c].shape[0] for c in cloudy_coords]),
            coords=[table[c] for c in cloudy_coords]
        )
        table['cloudy_box_amf'].attrs['long_name'] = 'NO2 box air mass factors'
        table['cloudy_box_amf'].attrs['units'] = '1'

        # add a data variable for uncertainty if this table will be processed with the MC engine
        if engine == 'mc':
            table['cloudy_box_amf_std'] = xr.DataArray(
                data=np.zeros([table[c].shape[0] for c in cloudy_coords]),
                coords=[table[c] for c in cloudy_coords]
            )
            table['cloudy_box_amf_std'].attrs['long_name'] = 'NO2 box air mass factor standard deviations'
            table['cloudy_box_amf_std'].attrs['units'] = '1'

        # add data variable for total AMFs if requested
        cloudy_coords.remove('box_altitude')
        if total_sum:
            table['cloudy_amf_box_sum'] = xr.DataArray(
                data=np.zeros([table[c].shape[0] for c in cloudy_coords]),
                coords=[table[c] for c in cloudy_coords]
            )
        if total_remove:
            table['cloudy_amf_no_NO2'] = xr.DataArray(
                data=np.zeros([table[c].shape[0] for c in cloudy_coords]),
                coords=[table[c] for c in cloudy_coords]
            )
        del cloudy_coords
    return table


def create_table(filename, copy=None, copy_options=None, clear_sky=False, cloudy_sky=False, atmosphere=None, total_sum=False, total_remove=False, engine=None, **kwargs):
    # create the base table with no options, either with the new table dialog or by copying an existing table
    # if an existing table is copied, keep the atmosphere vars for now (will be overwritten if an atmosphere is specified)
    if copy is None:
        fresh_table = new_table(clear_sky, cloudy_sky, total_sum, total_remove, engine, **kwargs)
    else:
        with xr.open_dataset(copy).load() as old_table:
            fresh_table = old_table.copy()
        # if we are copying an MC table with many trials, we just want to copy 1 of them
        if 'trials' in fresh_table.dims:
            fresh_table = fresh_table.sel(trials=0)
            fresh_table = fresh_table.drop(['date_created', 'version'])
        for key in fresh_table.data_vars:
            if all([a not in fresh_table[key].attrs for a in ['HSP_atmosphere', 'static_atmosphere']]):
                fresh_table[key].values = np.zeros(fresh_table[key].shape)
            if 'version' in fresh_table[key].attrs:
                del fresh_table[key].attrs['version']
        # remove options
        options = []
        for key in fresh_table.data_vars:
            if 'sasktran_option' in fresh_table[key].attrs:
                options.append(key)
        fresh_table = fresh_table.drop(options)
        # remove attributes associated with a table in progress or completed
        for key in ['date_created', 'progress_file', 'radiative_transfer_engine', 'version']:
            if key in fresh_table.attrs:
                del fresh_table.attrs[key]

    # copy options from another table if specified
    # set the permitted engine attribute if specified.
    if copy_options is not None:
        with xr.open_dataset(copy_options).load() as options_table:
            keys = list(fresh_table.variables.keys())
            for key in keys:
                if 'sasktran_option' in fresh_table[key].attrs:
                    del fresh_table[key]
            for key in options_table.data_vars:
                if 'sasktran_option' in options_table[key].attrs:
                    fresh_table[key] = options_table[key].drop(list(options_table[key].coords.keys()))
            for key in options_table.coords:
                if 'sasktran_option' in options_table[key].attrs:
                    fresh_table.coords[key] = options_table[key].drop(list(options_table[key].coords.keys()))
            if 'permitted_engine' in options_table.attrs:
                fresh_table.attrs['permitted_engine'] = options_table.attrs['permitted_engine']
            options_table.close()
    else:
        fresh_table.attrs['permitted_engine'] = engine

    # replace existing atmosphere variables (if any) with specified atmosphere
    merge = True
    if atmosphere is not None:
        with xr.open_dataset(atmosphere) as atmo:
            is_hsp_atmosphere(atmo)
            if 'atmosphere' in fresh_table.attrs:
                del fresh_table.attrs['atmosphere']
                remove = [var for var in fresh_table.variables if 'HSP_atmosphere' in fresh_table[var].attrs]
                remove += [var for var in fresh_table.variables if 'static_atmosphere' in fresh_table[var].attrs]
                for var in remove:
                    del fresh_table[var]
    else:
        if 'atmosphere' not in fresh_table.attrs:
            print('no atmosphere was specified so the default hsp atmosphere was used')
            atmo = default_hsp_atmosphere()
        else:
            atmo = None
            merge = False
    if merge:
        attrs = atmo.attrs
        attrs.update(fresh_table.attrs)
        fresh_table = xr.merge([fresh_table, atmo])
        fresh_table.attrs = attrs

    fresh_table.to_netcdf(filename)
    return fresh_table


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description="Generate an new empty AMF table.")
    parser.add_argument('filename', type=str, help='path to the location the table will be saved to.')
    parser.add_argument('--box-altitude', nargs='+', type=str, help='table parameter: center altitudes of AMF layers (m)')
    parser.add_argument('--surface-albedo', nargs='+', type=str, help='table parameter: surface albedos')
    parser.add_argument('--surface-pressure', nargs='+', type=str, help='table parameter: surface pressures (kPa)')
    parser.add_argument('--cloud_top_pressure', nargs='+', type=str, help='table parameter: cloud top pressures (kPa)')
    parser.add_argument('--cloud_albedo', nargs='+', type=str, help='table parameter: cloud albedo')
    parser.add_argument('--solar-zenith', nargs='+', type=str, help='table parameter: solar zenith angle (deg)')
    parser.add_argument('--viewing-zenith', nargs='+', type=str, help='table parameter: viewing zenith angle (deg)')
    parser.add_argument('--azimuth-difference', nargs='+', type=str, help='table parameter: difference between solar and viewing azimuth angles (deg)')
    parser.add_argument('-c', '--copy', type=str, help='table to copy coords and data_vars from')
    parser.add_argument('--copy-options', type=str, help='table to copy options (including permitted engines) from')
    parser.add_argument('--clear-sky', action='store_true')
    parser.add_argument('--cloudy-sky', action='store_true')
    parser.add_argument('-a', '--atmosphere', type=str, help='path to atmosphere netcdf file produced by skamf.atmosphere')
    parser.add_argument('-e', '--engine', type=str, choices=['hr', 'mc', 'do', 'disco'], default='do', help='engine that will be allowed to process the table - defaults to do')
    parser.add_argument('--total-sum', action='store_true', help='set to calculate the total AMF by summing the box-AMFs')
    parser.add_argument('--total-remove', action='store_true', help='set to estimate the total AMF by comparing radiances with the absorber removed')

    args = vars(parser.parse_args())

    if args['copy'] is None and args['clear_sky'] is False and args['cloudy_sky'] is False:
        parser.error("At least one of '--clear-sky' and '--cloudy-sky' must be selected")

    if args['engine'] is None and args['copy_options'] is None:
        parser.error("If '--copy-options' is not selected then '--engine' must be set")

    table = create_table(**args)
    print('Created empty AMF table:')
    print(table)
